#ifndef __VCPU_H__
#define __VCPU_H__


struct trap_bounce {
    unsigned long  error_code;
    unsigned short flags; /* TBF_ */
    unsigned short cs;
    unsigned long  eip;
};


#define VGCF_I387_VALID (1<<0)
#define VGCF_HVM_GUEST  (1<<1)
#define VGCF_IN_KERNEL  (1<<2)

/*
 * The following is all CPU context. Note that the fpu_ctxt block is filled 
 * in by FXSAVE if the CPU has feature FXSR; otherwise FSAVE is used.
 */
struct vcpu_guest_context {
	/* FPU registers come first so they can be aligned for FXSAVE/FXRSTOR. */
	struct { char x[512]; } fpu_ctxt;       /* User-level FPU registers     */
	unsigned long flags;                    /* VGCF_* flags                 */
	cpu_user_regs_t user_regs;              /* User-level CPU registers     */
	struct trap_info trap_ctxt[256];        /* Virtual IDT                  */
	unsigned long ldt_base, ldt_ents;       /* LDT (linear address, # ents) */
	unsigned long gdt_frames[16], gdt_ents; /* GDT (machine frames, # ents) */
	unsigned long kernel_ss, kernel_sp;     /* Virtual TSS (only SS1/SP1)   */
	unsigned long ctrlreg[8];               /* CR0-CR7 (control registers)  */
	unsigned long debugreg[8];              /* DB0-DB7 (debug registers)    */
	unsigned long event_callback_eip;
	unsigned long failsafe_callback_eip;
	unsigned long syscall_callback_eip;
	unsigned long vm_assist;                /* VMASST_TYPE_* bitmap */
	/* Segment base addresses. */
	uint64_t      fs_base;
	uint64_t      gs_base_kernel;
	uint64_t      gs_base_user;
};


struct arch_vcpu_info {
	unsigned long cr2;
	unsigned long pad; /* sizeof(vcpu_info_t) == 64 */
};

struct vcpu_info {
	/*
	 * 'evtchn_upcall_pending' is written non-zero by Xen to indicate
	 * a pending notification for a particular VCPU. It is then cleared 
	 * by the guest OS /before/ checking for pending work, thus avoiding
	 * a set-and-check race. Note that the mask is only accessed by Xen
	 * on the CPU that is currently hosting the VCPU. This means that the
	 * pending and mask flags can be updated by the guest without special
	 * synchronisation (i.e., no need for the x86 LOCK prefix).
	 * This may seem suboptimal because if the pending flag is set by
	 * a different CPU then an IPI may be scheduled even when the mask
	 * is set. However, note:
	 *  1. The task of 'interrupt holdoff' is covered by the per-event-
	 *     channel mask bits. A 'noisy' event that is continually being
	 *     triggered can be masked at source at this very precise
	 *     granularity.
	 *  2. The main purpose of the per-VCPU mask is therefore to restrict
	 *     reentrant execution: whether for concurrency control, or to
	 *     prevent unbounded stack usage. Whatever the purpose, we expect
	 *     that the mask will be asserted only for short periods at a time,
	 *     and so the likelihood of a 'spurious' IPI is suitably small.
	 * The mask is read before making an event upcall to the guest: a
	 * non-zero mask therefore guarantees that the VCPU will not receive
	 * an upcall activation. The mask is cleared when the VCPU requests
	 * to block: this avoids wakeup-waiting races.
	 */
	uint8_t evtchn_upcall_pending;
	uint8_t evtchn_upcall_mask;
	unsigned long evtchn_pending_sel;
	arch_vcpu_info_t arch;
	vcpu_time_info_t time;
}; /* 64 bytes (x86) */


struct hvm_vcpu {
	unsigned long   ioflags;
	struct mmio_op  mmio_op;
	struct vlapic   *vlapic;
	
        struct arch_svm_struct svm;
};

struct arch_vcpu {
	/* Needs 16-byte aligment for FXSAVE/FXRSTOR. */
	struct vcpu_guest_context guest_context
	__attribute__((__aligned__(16)));

	unsigned long      flags; /* TF_ */
	
	/* Bounce information for propagating an exception to guest OS. */
	struct trap_bounce trap_bounce;
	
	/* I/O-port access bitmap. */
	u8 *iobmp;        /* Guest kernel virtual address of the bitmap. */
	int iobmp_limit;  /* Number of ports represented in the bitmap.  */
	int iopl;         /* Current IOPL for this VCPU. */

	/* Virtual Machine Extensions */
	struct hvm_vcpu hvm_vcpu;

	/*
	 * Every domain has a L1 pagetable of its own. Per-domain mappings
	 * are put in this table (eg. the current GDT is mapped here).
	 */
	l1_pgentry_t *perdomain_ptes;
	
	pagetable_t  guest_table_user;      /* x86/64: user-space pagetable. */
	pagetable_t  guest_table;           /* (MA) guest notion of cr3 */
	pagetable_t  shadow_table;          /* (MA) shadow of guest */
	pagetable_t  monitor_table;         /* (MA) used in hypervisor */

	l2_pgentry_t *guest_vtable;         /* virtual address of pagetable */
	l2_pgentry_t *shadow_vtable;        /* virtual address of shadow_table */
	l2_pgentry_t *monitor_vtable;		/* virtual address of monitor_table */
	l1_pgentry_t *hl2_vtable;			/* virtual address of hl2_table */
	
	l3_pgentry_t *guest_vl3table;
	l4_pgentry_t *guest_vl4table;

	unsigned long monitor_shadow_ref;

	/* Current LDT details. */
	unsigned long shadow_ldt_mapcnt;
};

struct vcpu {
	struct domain		*domain;
	struct vcpu_info	*vcpu_info;

	
	
	struct timer     timer;         /* one-shot timer for timeout values */
	unsigned long    sleep_tick;    /* tick at which this vcpu started sleep */
	
	struct timer     poll_timer;    /* timeout for SCHEDOP_poll */
	
	void            *sched_priv;    /* scheduler-specific data */
	
	struct vcpu_runstate_info runstate;
	struct vcpu_runstate_info *runstate_guest; /* guest address */
	
	unsigned long    vcpu_flags;
	
	u16              virq_to_evtchn[NR_VIRQS];
	
	atomic_t         pausecnt;
	
	unsigned long    nmi_addr;      /* NMI callback address. */
	
	struct arch_vcpu arch;
};




/* Anonymous union includes both 32- and 64-bit names (e.g., eax/rax). */
#define __DECL_REG(name) union { uint64_t r ## name, e ## name; }


struct cpu_user_regs {
	uint64_t r15;
	uint64_t r14;
	uint64_t r13;
	uint64_t r12;
	__DECL_REG(bp);
	__DECL_REG(bx);
	uint64_t r11;
	uint64_t r10;
	uint64_t r9;
	uint64_t r8;
	__DECL_REG(ax);
	__DECL_REG(cx);
	__DECL_REG(dx);
	__DECL_REG(si);
	__DECL_REG(di);
	uint32_t error_code;    /* private */
	uint32_t entry_vector;  /* private */
	__DECL_REG(ip);
	uint16_t cs, _pad0[1];
	uint8_t  saved_upcall_mask;
	uint8_t  _pad1[3];
	__DECL_REG(flags);      /* rflags.IF == !saved_upcall_mask */
	__DECL_REG(sp);
	uint16_t ss, _pad2[3];
	uint16_t es, _pad3[3];
	uint16_t ds, _pad4[3];
	uint16_t fs, _pad5[3]; /* Non-zero => takes precedence over fs_base.     */
	uint16_t gs, _pad6[3]; /* Non-zero => takes precedence over gs_base_usr. */
};

struct cpu_info {
    struct cpu_user_regs guest_cpu_user_regs;
    struct vcpu  *current_ed;
};


static inline struct cpu_info *get_cpu_info(void)
{
    struct cpu_info *cpu_info;
    __asm__ ( "and %%"__OP"sp,%0; or %2,%0"
              : "=r" (cpu_info)
              : "0" (~(STACK_SIZE-1)), "i" (STACK_SIZE-sizeof(struct cpu_info))
        );
    return cpu_info;
}

#define guest_cpu_user_regs() (&get_cpu_info()->guest_cpu_user_regs)

#define reset_stack_and_jump(__fn)              \
    __asm__ __volatile__ (                      \
        "mov %0,%%rsp; jmp "STR(__fn)      \
        : : "r" ( guest_cpu_user_regs ( ) ) : "memory" )


#endif /* __VCPU_H__ */
