#define __pge_off()                                                     \
    do {                                                                \
        __asm__ __volatile__(                                           \
            "mov %0, %%cr4;  # turn off PGE     "                       \
            : : "r" (mmu_cr4_features & ~X86_CR4_PGE) );                \
        } while ( 0 )

#define __pge_on()                                                      \
    do {                                                                \
        __asm__ __volatile__(                                           \
            "mov %0, %%cr4;  # turn off PGE     "                       \
            : : "r" (mmu_cr4_features) );                               \
    } while ( 0 )


/* Write pagetable base and implicitly tick the tlbflush clock. */
extern void write_cr3(unsigned long cr3);

#define local_flush_tlb()                                         \
    do {                                                          \
        unsigned long cr3 = read_cr3();                           \
        write_cr3(cr3);                                           \
    } while ( 0 )


#define local_flush_tlb_pge()                                     \
    do {                                                          \
        __pge_off();                                              \
        local_flush_tlb();                                        \
        __pge_on();                                               \
    } while ( 0 )





#define __flush_tlb()							\
	do {								\
		unsigned long tmpreg;					\
									\
		__asm__ __volatile__(					\
			"movq %%cr3, %0;  # flush TLB \n"		\
			"movq %0, %%cr3;              \n"		\
			: "=r" (tmpreg)					\
			:: "memory");					\
	} while (0)

/*
 * Global pages have to be flushed a bit differently. Not a real
 * performance problem because this does not happen often.
 */
#define __flush_tlb_global()						\
	do {								\
		unsigned long tmpreg, cr4, cr4_orig;			\
									\
		__asm__ __volatile__(					\
			"movq %%cr4, %2;  # turn off PGE     \n"	\
			"movq %2, %1;                        \n"	\
			"andq %3, %1;                        \n"	\
			"movq %1, %%cr4;                     \n"	\
			"movq %%cr3, %0;  # flush TLB        \n"	\
			"movq %0, %%cr3;                     \n"	\
			"movq %2, %%cr4;  # turn PGE back on \n"	\
			: "=&r" (tmpreg), "=&r" (cr4), "=&r" (cr4_orig)	\
			: "i" (~X86_CR4_PGE)				\
			: "memory");					\
	} while (0)

#define __flush_tlb_all() __flush_tlb_global()

#define __flush_tlb_one(addr) \
	__asm__ __volatile__("invlpg %0": :"m" (*(char *) addr))





/******************************************************************************
 * flushtlb.c
 * 
 * TLB flushes are timestamped using a global virtual 'clock' which ticks
 * on any TLB flush on any processor.
 * 
 * Copyright (c) 2003-2004, K A Fraser
 */

#include <xen/config.h>
#include <xen/sched.h>
#include <xen/softirq.h>
#include <asm/flushtlb.h>

/* Debug builds: Wrap frequently to stress-test the wrap logic. */
#ifdef NDEBUG
#define WRAP_MASK (0xFFFFFFFFU)
#else
#define WRAP_MASK (0x000003FFU)
#endif

u32 tlbflush_clock = 1U;
u32 tlbflush_time[NR_CPUS];

void write_cr3(unsigned long cr3)
{
	u32 t, t1, t2;
	unsigned long flags;

	/* This non-reentrant function is sometimes called in interrupt context. */
	local_irq_save(flags);

	/*
	 * STEP 1. Increment the virtual clock *before* flushing the TLB.
	 *         If we do it after, we race other CPUs invalidating PTEs.
	 *         (e.g., a page invalidated after the flush might get the old 
	 *          timestamp, but this CPU can speculatively fetch the mapping
	 *          into its TLB after the flush but before inc'ing the clock).
	 */

	t = tlbflush_clock;
	do {
		t1 = t2 = t;
		/* Clock wrapped: someone else is leading a global TLB shootdown. */
		if ( unlikely(t1 == 0) )
			goto skip_clocktick;
		t2 = (t + 1) & WRAP_MASK;
	}
	while ( unlikely((t = cmpxchg(&tlbflush_clock, t1, t2)) != t1) );

	/* Clock wrapped: we will lead a global TLB shootdown. */
	if ( unlikely(t2 == 0) )
		raise_softirq(NEW_TLBFLUSH_CLOCK_PERIOD_SOFTIRQ);

	/*
	 * STEP 2. Update %CR3, thereby flushing the TLB.
	 */

skip_clocktick:
	__asm__ __volatile__ ( "mov %0, %%cr3" : : "r" (cr3) : "memory" );

	/*
	 * STEP 3. Update this CPU's timestamp. Note that this happens *after*
	 *         flushing the TLB, as otherwise we can race a NEED_FLUSH() test
	 *         on another CPU. (e.g., other CPU sees the updated CPU stamp and
	 *         so does not force a synchronous TLB flush, but the flush in this
	 *         function hasn't yet occurred and so the TLB might be stale).
	 *         The ordering would only actually matter if this function were
	 *         interruptible, and something that abuses the stale mapping could
	 *         exist in an interrupt handler. In fact neither of these is the
	 *         case, so really we are being ultra paranoid.
	 */

	tlbflush_time[smp_processor_id()] = t2;

	local_irq_restore(flags);
}










#define L1_PAGETABLE_SHIFT      12
#define L2_PAGETABLE_SHIFT      21
#define L3_PAGETABLE_SHIFT      30
#define L4_PAGETABLE_SHIFT      39
#define PAGE_SHIFT              L1_PAGETABLE_SHIFT
#define ROOT_PAGETABLE_SHIFT    L4_PAGETABLE_SHIFT

#define PAGETABLE_ORDER         9
#define L1_PAGETABLE_ENTRIES    (1<<PAGETABLE_ORDER)
#define L2_PAGETABLE_ENTRIES    (1<<PAGETABLE_ORDER)
#define L3_PAGETABLE_ENTRIES    (1<<PAGETABLE_ORDER)
#define L4_PAGETABLE_ENTRIES    (1<<PAGETABLE_ORDER)
#define ROOT_PAGETABLE_ENTRIES  L4_PAGETABLE_ENTRIES

#define __PAGE_OFFSET           (0xFFFF830000000000)

/* These are architectural limits. Current CPUs support only 40-bit phys. */
#define PADDR_BITS              52
#define VADDR_BITS              48
#define PADDR_MASK              ((1UL << PADDR_BITS)-1)
#define VADDR_MASK              ((1UL << VADDR_BITS)-1)

/* Given a virtual address, get an entry offset into a linear page table. */
#define l1_linear_offset(_a) (((_a) & VADDR_MASK) >> L1_PAGETABLE_SHIFT)
#define l2_linear_offset(_a) (((_a) & VADDR_MASK) >> L2_PAGETABLE_SHIFT)

/* Given a virtual address, get an entry offset into a page table. */
#define l1_table_offset(a)         \
    (((a) >> L1_PAGETABLE_SHIFT) & (L1_PAGETABLE_ENTRIES - 1))
#define l2_table_offset(a)         \
    (((a) >> L2_PAGETABLE_SHIFT) & (L2_PAGETABLE_ENTRIES - 1))
#define l3_table_offset(a)         \
    (((a) >> L3_PAGETABLE_SHIFT) & (L3_PAGETABLE_ENTRIES - 1))
#define l4_table_offset(a)         \
    (((a) >> L4_PAGETABLE_SHIFT) & (L4_PAGETABLE_ENTRIES - 1))


#define clear_page(_p)      memset((void *)(_p), 0, PAGE_SIZE)


/* Construct a pte from a pfn and access flags. */
#define l1e_from_pfn(pfn, flags)   \
    ((l1_pgentry_t) { ((intpte_t)(pfn) << PAGE_SHIFT) | put_pte_flags(flags) })
#define l2e_from_pfn(pfn, flags)   \
    ((l2_pgentry_t) { ((intpte_t)(pfn) << PAGE_SHIFT) | put_pte_flags(flags) })
#define l3e_from_pfn(pfn, flags)   \
    ((l3_pgentry_t) { ((intpte_t)(pfn) << PAGE_SHIFT) | put_pte_flags(flags) })
#define l4e_from_pfn(pfn, flags)   \
    ((l4_pgentry_t) { ((intpte_t)(pfn) << PAGE_SHIFT) | put_pte_flags(flags) })

/* Construct a pte from a physical address and access flags. */
#ifndef __ASSEMBLY__
static inline l1_pgentry_t l1e_from_paddr(paddr_t pa, unsigned int flags)
{
    ASSERT((pa & ~(PADDR_MASK & PAGE_MASK)) == 0);
    return (l1_pgentry_t) { pa | put_pte_flags(flags) };
}
static inline l2_pgentry_t l2e_from_paddr(paddr_t pa, unsigned int flags)
{
    ASSERT((pa & ~(PADDR_MASK & PAGE_MASK)) == 0);
    return (l2_pgentry_t) { pa | put_pte_flags(flags) };
}
#if CONFIG_PAGING_LEVELS >= 3
static inline l3_pgentry_t l3e_from_paddr(paddr_t pa, unsigned int flags)
{
    ASSERT((pa & ~(PADDR_MASK & PAGE_MASK)) == 0);
    return (l3_pgentry_t) { pa | put_pte_flags(flags) };
}
#endif
#if CONFIG_PAGING_LEVELS >= 4
static inline l4_pgentry_t l4e_from_paddr(paddr_t pa, unsigned int flags)
{
    ASSERT((pa & ~(PADDR_MASK & PAGE_MASK)) == 0);
    return (l4_pgentry_t) { pa | put_pte_flags(flags) };
}
#endif
#endif /* !__ASSEMBLY__ */


/* Convert between machine frame numbers and page-info structures. */
#define mfn_to_page(mfn)    (frame_table + (mfn))

/* Convert between Xen-heap virtual addresses and page-info structures. */
#define virt_to_page(va)    (frame_table + (__pa(va) >> PAGE_SHIFT))
#define page_to_virt(pg)    (maddr_to_virt(page_to_maddr(pg)))


struct page_info *
alloc_xen_pagetable ( void )
{
	extern int early_boot;
	unsigned long pfn;

	if ( ! early_boot ) {
		return alloc_domheap_page ( NULL );
	}

	pfn = alloc_boot_pages (1 /* nr_pfns */, 1 /* pfn_align */ );
	return mfn_to_page ( pfn );
}

void
free_xen_pagetable ( struct page_info *pg )
{
	free_domheap_page(pg);
}

l2_pgentry_t *
virt_to_xen_l2e ( unsigned long v )
{
	l4_pgentry_t *pl4e;
	l3_pgentry_t *pl3e;
	l2_pgentry_t *pl2e;
	
	/* Look up Page-Map Level-4 Table.  */
	pl4e = & idle_pg_table [ l4_table_offset ( v ) ];
	if ( !(l4e_get_flags(*pl4e) & _PAGE_PRESENT) )
	{
		pl3e = page_to_virt(alloc_xen_pagetable());
		clear_page(pl3e);
		*pl4e = l4e_from_paddr(__pa(pl3e), __PAGE_HYPERVISOR); /*  entry äƤ */
	}

	/* Look up Page-Directory Pointer Table */
    
	pl3e = l4e_to_l3e(*pl4e) + l3_table_offset(v);
	if ( !(l3e_get_flags(*pl3e) & _PAGE_PRESENT) )
	{
		pl2e = page_to_virt(alloc_xen_pagetable());
		clear_page(pl2e);
		*pl3e = l3e_from_paddr(__pa(pl2e), __PAGE_HYPERVISOR);
	}
    
	/* Look up Page-Directory Table */

	pl2e = l3e_to_l2e(*pl3e) + l2_table_offset(v);
	return pl2e;
}

void __init
paging_init ( void )
{
	unsigned long i, mpt_size;
	l3_pgentry_t *l3_ro_mpt;
	l2_pgentry_t *l2_ro_mpt;
	struct page_info *pg;

	idle_vcpu[0]->arch.monitor_table = mk_pagetable ( PHYS ( idle_pg_table ) );

	/* Create user-accessible L2 directory to map the MPT for guests. */
	l3_ro_mpt = alloc_xenheap_page ( );
	clear_page ( l3_ro_mpt );
	idle_pg_table [ l4_table_offset ( RO_MPT_VIRT_START ) ] = l4e_from_page (  virt_to_page ( l3_ro_mpt ), __PAGE_HYPERVISOR | _PAGE_USER );
	l2_ro_mpt = alloc_xenheap_page ();
	clear_page ( l2_ro_mpt );
	l3_ro_mpt [ l3_table_offset ( RO_MPT_VIRT_START ) ] = l3e_from_page ( virt_to_page ( l2_ro_mpt ), __PAGE_HYPERVISOR | _PAGE_USER );
	l2_ro_mpt += l2_table_offset ( RO_MPT_VIRT_START );

	/*
	 * Allocate and map the machine-to-phys table.
	 * This also ensures L3 is present for fixmaps.
	 */
	mpt_size  = (max_page * BYTES_PER_LONG) + (1UL << L2_PAGETABLE_SHIFT) - 1;
	mpt_size &= ~((1UL << L2_PAGETABLE_SHIFT) - 1UL);
	for ( i = 0; i < (mpt_size >> L2_PAGETABLE_SHIFT); i++ ) { 

		if ( (pg = alloc_domheap_pages(NULL, PAGETABLE_ORDER, 0)) == NULL ) {
			panic("Not enough memory for m2p table\n");
		}

		map_pages_to_xen(
			RDWR_MPT_VIRT_START + (i << L2_PAGETABLE_SHIFT), page_to_mfn(pg), 
			1UL << PAGETABLE_ORDER,
			PAGE_HYPERVISOR);
		memset((void *)(RDWR_MPT_VIRT_START + (i << L2_PAGETABLE_SHIFT)), 0x55,
		       1UL << L2_PAGETABLE_SHIFT);
		*l2_ro_mpt++ = l2e_from_page(
			pg, _PAGE_GLOBAL|_PAGE_PSE|_PAGE_USER|_PAGE_PRESENT);
		BUG_ON(((unsigned long)l2_ro_mpt & ~PAGE_MASK) == 0);
	}

	/* Set up linear page table mapping. */
	idle_pg_table[l4_table_offset(LINEAR_PT_VIRT_START)] = l4e_from_paddr(__pa(idle_pg_table), __PAGE_HYPERVISOR);

	/* Install per-domain mappings for idle domain. */
	idle_pg_table[l4_table_offset(PERDOMAIN_VIRT_START)] = l4e_from_page ( virt_to_page(idle_vcpu[0]->domain->arch.mm_perdomain_l3), __PAGE_HYPERVISOR);
}

void __init zap_low_mappings(void)
{
	idle_pg_table[0] = l4e_empty();
	flush_tlb_all_pge();
}

void subarch_init_memory(void)
{
	unsigned long i, v, m2p_start_mfn;
	l3_pgentry_t l3e;
	l2_pgentry_t l2e;

	/*
	 * We are rather picky about the layout of 'struct page_info'. The
	 * count_info and domain fields must be adjacent, as we perform atomic
	 * 64-bit operations on them.
	 */
	if ( ((offsetof(struct page_info, u.inuse._domain) != 
	       (offsetof(struct page_info, count_info) + sizeof(u32)))) ||
	     ((offsetof(struct page_info, count_info) & 7) != 0) ||
	     (sizeof(struct page_info) !=
	      (32 + BITS_TO_LONGS(NR_CPUS)*sizeof(long))) )
	{
		printk("Weird page_info layout (%ld,%ld,%ld,%ld)\n",
		       offsetof(struct page_info, count_info),
		       offsetof(struct page_info, u.inuse._domain),
		       sizeof(struct page_info),
		       32 + BITS_TO_LONGS(NR_CPUS)*sizeof(long));
		for ( ; ; ) ;
	}

	/* M2P table is mappable read-only by privileged domains. */
	for ( v  = RDWR_MPT_VIRT_START; 
	      v != RDWR_MPT_VIRT_END;
	      v += 1 << L2_PAGETABLE_SHIFT )
	{
		l3e = l4e_to_l3e(idle_pg_table[l4_table_offset(v)])[
			l3_table_offset(v)];
		if ( !(l3e_get_flags(l3e) & _PAGE_PRESENT) )
			continue;
		l2e = l3e_to_l2e(l3e)[l2_table_offset(v)];
		if ( !(l2e_get_flags(l2e) & _PAGE_PRESENT) )
			continue;
		m2p_start_mfn = l2e_get_pfn(l2e);

		for ( i = 0; i < L1_PAGETABLE_ENTRIES; i++ )
		{
			struct page_info *page = mfn_to_page(m2p_start_mfn + i);
			share_xen_page_with_privileged_guests(page, XENSHARE_readonly);
		}
	}
}

long subarch_memory_op(int op, GUEST_HANDLE(void) arg)
{
	struct xen_machphys_mfn_list xmml;
	l3_pgentry_t l3e;
	l2_pgentry_t l2e;
	unsigned long mfn, v;
	unsigned int i;
	long rc = 0;

	switch ( op )
	{
	case XENMEM_machphys_mfn_list:
		if ( copy_from_guest(&xmml, arg, 1) )
			return -EFAULT;

		for ( i = 0, v = RDWR_MPT_VIRT_START;
		      (i != xmml.max_extents) && (v != RDWR_MPT_VIRT_END);
		      i++, v += 1 << 21 )
		{
			l3e = l4e_to_l3e(idle_pg_table[l4_table_offset(v)])[
				l3_table_offset(v)];
			if ( !(l3e_get_flags(l3e) & _PAGE_PRESENT) )
				break;
			l2e = l3e_to_l2e(l3e)[l2_table_offset(v)];
			if ( !(l2e_get_flags(l2e) & _PAGE_PRESENT) )
				break;
			mfn = l2e_get_pfn(l2e) + l1_table_offset(v);
			if ( copy_to_guest_offset(xmml.extent_start, i, &mfn, 1) )
				return -EFAULT;
		}

		xmml.nr_extents = i;
		if ( copy_to_guest(arg, &xmml, 1) )
			return -EFAULT;

		break;

	default:
		rc = -ENOSYS;
		break;
	}

	return rc;
}

long do_stack_switch(unsigned long ss, unsigned long esp)
{
	fixup_guest_stack_selector(ss);
	current->arch.guest_context.kernel_ss = ss;
	current->arch.guest_context.kernel_sp = esp;
	return 0;
}

long do_set_segment_base(unsigned int which, unsigned long base)
{
	struct vcpu *v = current;
	long ret = 0;

	switch ( which )
	{
	case SEGBASE_FS:
		if ( wrmsr_safe(MSR_FS_BASE, base, base>>32) )
			ret = -EFAULT;
		else
			v->arch.guest_context.fs_base = base;
		break;

	case SEGBASE_GS_USER:
		if ( wrmsr_safe(MSR_SHADOW_GS_BASE, base, base>>32) )
			ret = -EFAULT;
		else
			v->arch.guest_context.gs_base_user = base;
		break;

	case SEGBASE_GS_KERNEL:
		if ( wrmsr_safe(MSR_GS_BASE, base, base>>32) )
			ret = -EFAULT;
		else
			v->arch.guest_context.gs_base_kernel = base;
		break;

	case SEGBASE_GS_USER_SEL:
		__asm__ __volatile__ (
			"     swapgs              \n"
			"1:   movl %k0,%%gs       \n"
			"    "safe_swapgs"        \n"
			".section .fixup,\"ax\"   \n"
			"2:   xorl %k0,%k0        \n"
			"     jmp  1b             \n"
			".previous                \n"
			".section __ex_table,\"a\"\n"
			"    .align 8             \n"
			"    .quad 1b,2b          \n"
			".previous                  "
			: : "r" (base&0xffff) );
		break;

	default:
		ret = -EINVAL;
		break;
	}

	return ret;
}


/* Returns TRUE if given descriptor is valid for GDT or LDT. */
int check_descriptor(struct desc_struct *d)
{
	u32 a = d->a, b = d->b;
	u16 cs;

	/* A not-present descriptor will always fault, so is safe. */
	if ( !(b & _SEGMENT_P) ) 
		goto good;

	/* Check and fix up the DPL. */
	if ( (b & _SEGMENT_DPL) < (GUEST_KERNEL_RPL << 13) )
		d->b = b = (b & ~_SEGMENT_DPL) | (GUEST_KERNEL_RPL << 13);

	/* All code and data segments are okay. No base/limit checking. */
	if ( (b & _SEGMENT_S) )
		goto good;

	/* Invalid type 0 is harmless. It is used for 2nd half of a call gate. */
	if ( (b & _SEGMENT_TYPE) == 0x000 )
		goto good;

	/* Everything but a call gate is discarded here. */
	if ( (b & _SEGMENT_TYPE) != 0xc00 )
		goto bad;

	/* Validate and fix up the target code selector. */
	cs = a >> 16;
	fixup_guest_code_selector(cs);
	if ( !guest_gate_selector_okay(cs) )
		goto bad;
	a = d->a = (d->a & 0xffffU) | (cs << 16);

	/* Reserved bits must be zero. */
	if ( (b & 0xe0) != 0 )
		goto bad;
        
good:
	return 1;
bad:
	return 0;
}

void memguard_guard_stack(void *p)
{
	p = (void *)((unsigned long)p + PAGE_SIZE);
	memguard_guard_range(p, 2 * PAGE_SIZE);
}

/*******************************************/

int 
map_pages_to_xen ( unsigned long virt, unsigned long mfn, unsigned long nr_mfns, unsigned long flags )
{
	l2_pgentry_t *pl2e, ol2e;
	l1_pgentry_t *pl1e, ol1e;
	unsigned int i;

	unsigned int map_small_pages = !! ( flags & MAP_SMALL_PAGES );
	flags &= ~MAP_SMALL_PAGES;

	while ( nr_mfns != 0 ) {
		pl2e = virt_to_xen_l2e ( virt );

		if ( ( ( ( ( virt >> PAGE_SHIFT ) | mfn ) & ( ( 1 << PAGETABLE_ORDER ) - 1 ) ) == 0 ) &&
		     ( nr_mfns >= ( 1 << PAGETABLE_ORDER ) ) &&
		     ! map_small_pages ) {

			/* Super-page mapping. */
			ol2e  = *pl2e;
			*pl2e = l2e_from_pfn(mfn, flags|_PAGE_PSE);

			if ( ( l2e_get_flags ( ol2e ) & _PAGE_PRESENT ) )
			{
				local_flush_tlb_pge ( ); // flush_tlb_global 

				if ( ! ( l2e_get_flags ( ol2e ) & _PAGE_PSE ) ) {
					free_xen_pagetable ( l2e_get_page ( *pl2e ) );
				}
			}

			virt    += 1UL << L2_PAGETABLE_SHIFT;
			mfn     += 1UL << PAGETABLE_ORDER;
			nr_mfns -= 1UL << PAGETABLE_ORDER;
		} else {
			/* Normal page mapping. */
			if ( ! ( l2e_get_flags ( *pl2e ) & _PAGE_PRESENT ) ) {
				/* note present */

				/* ƱͤνѽФ */
				pl1e = page_to_virt ( alloc_xen_pagetable ( ) );
				clear_page ( pl1e );
				*pl2e = l2e_from_page ( virt_to_page ( pl1e ), __PAGE_HYPERVISOR );

			} else if ( l2e_get_flags ( *pl2e ) & _PAGE_PSE ) {

				pl1e = page_to_virt ( alloc_xen_pagetable ( ) );
				for ( i = 0; i < L1_PAGETABLE_ENTRIES; i++ ) {
					pl1e[i] = l1e_from_pfn ( l2e_get_pfn ( *pl2e ) + i, l2e_get_flags ( *pl2e ) & ~_PAGE_PSE );
				}
				*pl2e = l2e_from_page ( virt_to_page ( pl1e ), __PAGE_HYPERVISOR);
				local_flush_tlb_pge (); // flush_tlb_global
			}

			pl1e  = l2e_to_l1e ( *pl2e ) + l1_table_offset ( virt );
			ol1e  = *pl1e;
			*pl1e = l1e_from_pfn ( mfn, flags );
			if ( ( l1e_get_flags ( ol1e ) & _PAGE_PRESENT ) ) {
				local_flush_tlb_one ( virt ); // flush_tlb_one()
			}

			virt    += 1UL << L1_PAGETABLE_SHIFT;
			mfn     += 1UL;
			nr_mfns -= 1UL;
		}
	}

	return 0;
}



static unsigned long *alloc_bitmap;
#define PAGES_PER_MAPWORD (sizeof(unsigned long) * 8)

#define allocated_in_map(_pn)                 \
( !! (alloc_bitmap[(_pn)/PAGES_PER_MAPWORD] & \
     (1UL<<((_pn)&(PAGES_PER_MAPWORD-1)))) )

/*
 * Hint regarding bitwise arithmetic in map_{alloc,free}:
 *  -(1<<n)  sets all bits >= n. 
 *  (1<<n)-1 sets all bits <  n.
 * Variable names in map_{alloc,free}:
 *  *_idx == Index into `alloc_bitmap' array.
 *  *_off == Bit offset within an element of the `alloc_bitmap' array.
 */

static void 
map_alloc ( unsigned long first_page, unsigned long nr_pages )
{
	unsigned long start_off, end_off, curr_idx, end_idx;
	
	curr_idx  = first_page / PAGES_PER_MAPWORD;
	start_off = first_page & ( PAGES_PER_MAPWORD - 1 );
	end_idx   = ( first_page + nr_pages ) / PAGES_PER_MAPWORD;
	end_off   = ( first_page + nr_pages ) & ( PAGES_PER_MAPWORD - 1 );
	
	if ( curr_idx == end_idx ) {
		alloc_bitmap [ curr_idx ] |= ( ( 1UL << end_off ) - 1 ) & - ( 1UL << start_off );
	} else  {
		alloc_bitmap [ curr_idx ] |= - ( 1UL << start_off );
		while ( ++curr_idx < end_idx ) {
			alloc_bitmap[ curr_idx ] = ~0UL;
		}
		alloc_bitmap [ curr_idx ] |= ( 1UL << end_off ) - 1;
	}
}

static void map_free(unsigned long first_page, unsigned long nr_pages)
{
    unsigned long start_off, end_off, curr_idx, end_idx;

#ifndef NDEBUG
    unsigned long i;
    /* Check that the block isn't already freed. */
    for ( i = 0; i < nr_pages; i++ )
        ASSERT(allocated_in_map(first_page + i));
#endif

    curr_idx  = first_page / PAGES_PER_MAPWORD;
    start_off = first_page & (PAGES_PER_MAPWORD-1);
    end_idx   = (first_page + nr_pages) / PAGES_PER_MAPWORD;
    end_off   = (first_page + nr_pages) & (PAGES_PER_MAPWORD-1);

    if ( curr_idx == end_idx )
    {
        alloc_bitmap[curr_idx] &= -(1UL<<end_off) | ((1UL<<start_off)-1);
    }
    else 
    {
        alloc_bitmap[curr_idx] &= (1UL<<start_off)-1;
        while ( ++curr_idx != end_idx ) alloc_bitmap[curr_idx] = 0;
        alloc_bitmap[curr_idx] &= -(1UL<<end_off);
    }
}

/*************************
 * BOOT-TIME ALLOCATOR
 */


/*************************
 * BINARY BUDDY ALLOCATOR
 */

#define MEMZONE_XEN 0
#define MEMZONE_DOM 1
#define MEMZONE_DMADOM 2
#define NR_ZONES    3

#define pfn_dom_zone_type(_pfn)                                 \
    (((_pfn) <= MAX_DMADOM_PFN) ? MEMZONE_DMADOM : MEMZONE_DOM)

static struct list_head heap[NR_ZONES][MAX_ORDER+1];

static unsigned long avail[NR_ZONES];

static spinlock_t heap_lock = SPIN_LOCK_UNLOCKED;


void end_boot_allocator(void)
{
	unsigned long i, j;
	int curr_free = 0, next_free = 0;

	memset(avail, 0, sizeof(avail));

	for ( i = 0; i < NR_ZONES; i++ )
		for ( j = 0; j <= MAX_ORDER; j++ )
			INIT_LIST_HEAD(&heap[i][j]);

	/* Pages that are free now go to the domain sub-allocator. */
	for ( i = 0; i < max_page; i++ )
	{
		curr_free = next_free;
		next_free = !allocated_in_map(i+1);
		if ( next_free )
			map_alloc(i+1, 1); /* prevent merging in free_heap_pages() */
		if ( curr_free )
			free_heap_pages(pfn_dom_zone_type(i), mfn_to_page(i), 0);
	}
}



}

/* Hand the specified arbitrary page range to the specified heap zone. */
void 
init_heap_pages ( unsigned int zone, struct page_info *pg, unsigned long nr_pages )
{
	unsigned long i;
	
	ASSERT(zone < NR_ZONES);

	for ( i = 0; i < nr_pages; i++ ) {
		free_heap_pages(zone, pg+i, 0);
	}
}


/* Allocate 2^@order contiguous pages. */
struct page_info *alloc_heap_pages(unsigned int zone, unsigned int order)
{
    int i;
    struct page_info *pg;

    ASSERT(zone < NR_ZONES);

    if ( unlikely(order > MAX_ORDER) )
        return NULL;

    spin_lock(&heap_lock);

    /* Find smallest order which can satisfy the request. */
    for ( i = order; i <= MAX_ORDER; i++ )
        if ( !list_empty(&heap[zone][i]) )
            goto found;

    /* No suitable memory blocks. Fail the request. */
    spin_unlock(&heap_lock);
    return NULL;

 found: 
    pg = list_entry(heap[zone][i].next, struct page_info, list);
    list_del(&pg->list);

    /* We may have to halve the chunk a number of times. */
    while ( i != order )
    {
        PFN_ORDER(pg) = --i;
        list_add_tail(&pg->list, &heap[zone][i]);
        pg += 1 << i;
    }
    
    map_alloc(page_to_mfn(pg), 1 << order);
    avail[zone] -= 1 << order;

    spin_unlock(&heap_lock);

    return pg;
}


/* Free 2^@order set of pages. */
void free_heap_pages(
    unsigned int zone, struct page_info *pg, unsigned int order)
{
    unsigned long mask;

    ASSERT(zone < NR_ZONES);
    ASSERT(order <= MAX_ORDER);

    spin_lock(&heap_lock);

    map_free(page_to_mfn(pg), 1 << order);
    avail[zone] += 1 << order;
    
    /* Merge chunks as far as possible. */
    while ( order < MAX_ORDER )
    {
        mask = 1 << order;

        if ( (page_to_mfn(pg) & mask) )
        {
            /* Merge with predecessor block? */
            if ( allocated_in_map(page_to_mfn(pg)-mask) ||
                 (PFN_ORDER(pg-mask) != order) )
                break;
            list_del(&(pg-mask)->list);
            pg -= mask;
        }
        else
        {
            /* Merge with successor block? */
            if ( allocated_in_map(page_to_mfn(pg)+mask) ||
                 (PFN_ORDER(pg+mask) != order) )
                break;
            list_del(&(pg+mask)->list);
        }
        
        order++;
    }

    PFN_ORDER(pg) = order;
    list_add_tail(&pg->list, &heap[zone][order]);

    spin_unlock(&heap_lock);
}


/*
 * Scrub all unallocated pages in all heap zones. This function is more
 * convoluted than appears necessary because we do not want to continuously
 * hold the lock or disable interrupts while scrubbing very large memory areas.
 */
void scrub_heap_pages(void)
{
    void *p;
    unsigned long pfn;
    int cpu = smp_processor_id();

    printk("Scrubbing Free RAM: ");

    for ( pfn = 0; pfn < max_page; pfn++ )
    {
        /* Every 100MB, print a progress dot. */
        if ( (pfn % ((100*1024*1024)/PAGE_SIZE)) == 0 )
            printk(".");

        if ( unlikely(softirq_pending(cpu)) )
            do_softirq();

        /* Quick lock-free check. */
        if ( allocated_in_map(pfn) )
            continue;

        spin_lock_irq(&heap_lock);

        /* Re-check page status with lock held. */
        if ( !allocated_in_map(pfn) )
        {
            if ( IS_XEN_HEAP_FRAME(mfn_to_page(pfn)) )
            {
                p = page_to_virt(mfn_to_page(pfn));
                memguard_unguard_range(p, PAGE_SIZE);
                clear_page(p);
                memguard_guard_range(p, PAGE_SIZE);
            }
            else
            {
                p = map_domain_page(pfn);
                clear_page(p);
                unmap_domain_page(p);
            }
        }

        spin_unlock_irq(&heap_lock);
    }

    printk("done.\n");
}






/*************************
 * XEN-HEAP SUB-ALLOCATOR
 */

void init_xenheap_pages(paddr_t ps, paddr_t pe)
{
    unsigned long flags;

    ps = round_pgup(ps);
    pe = round_pgdown(pe);
    if ( pe <= ps )
        return;

    memguard_guard_range(maddr_to_virt(ps), pe - ps);

    /*
     * Yuk! Ensure there is a one-page buffer between Xen and Dom zones, to
     * prevent merging of power-of-two blocks across the zone boundary.
     */
    if ( !IS_XEN_HEAP_FRAME(maddr_to_page(pe)) )
        pe -= PAGE_SIZE;

    local_irq_save(flags);
    init_heap_pages(MEMZONE_XEN, maddr_to_page(ps), (pe - ps) >> PAGE_SHIFT);
    local_irq_restore(flags);
}


void *alloc_xenheap_pages(unsigned int order)
{
    unsigned long flags;
    struct page_info *pg;
    int i;

    local_irq_save(flags);
    pg = alloc_heap_pages(MEMZONE_XEN, order);
    local_irq_restore(flags);

    if ( unlikely(pg == NULL) )
        goto no_memory;

    memguard_unguard_range(page_to_virt(pg), 1 << (order + PAGE_SHIFT));

    for ( i = 0; i < (1 << order); i++ )
    {
        pg[i].count_info        = 0;
        pg[i].u.inuse._domain   = 0;
        pg[i].u.inuse.type_info = 0;
    }

    return page_to_virt(pg);

 no_memory:
    printk("Cannot handle page request order %d!\n", order);
    return NULL;
}


void free_xenheap_pages(void *v, unsigned int order)
{
    unsigned long flags;

    if ( v == NULL )
        return;

    memguard_guard_range(v, 1 << (order + PAGE_SHIFT));    

    local_irq_save(flags);
    free_heap_pages(MEMZONE_XEN, virt_to_page(v), order);
    local_irq_restore(flags);
}


/* Frame table and its size in pages. */
struct page_info *frame_table;
unsigned long max_page;
unsigned long total_pages;

void __init init_frametable(void)
{
    unsigned long nr_pages, page_step, i, mfn;

    frame_table = (struct page_info *)FRAMETABLE_VIRT_START;

    nr_pages  = PFN_UP(max_page * sizeof(*frame_table));
    page_step = (1 << L2_PAGETABLE_SHIFT) >> PAGE_SHIFT;

    for ( i = 0; i < nr_pages; i += page_step )
    {
        mfn = alloc_boot_pages(min(nr_pages - i, page_step), page_step);
        if ( mfn == 0 )
            panic("Not enough memory for frame table\n");
        map_pages_to_xen(
            FRAMETABLE_VIRT_START + (i << PAGE_SHIFT),
            mfn, page_step, PAGE_HYPERVISOR);
    }

    memset(frame_table, 0, nr_pages << PAGE_SHIFT);
}



/*************************
 * DOMAIN-HEAP SUB-ALLOCATOR
 */

void init_domheap_pages(paddr_t ps, paddr_t pe)
{
    unsigned long s_tot, e_tot, s_dma, e_dma, s_nrm, e_nrm;

    ASSERT(!in_irq());

    s_tot = round_pgup(ps) >> PAGE_SHIFT;
    e_tot = round_pgdown(pe) >> PAGE_SHIFT;

    s_dma = min(s_tot, MAX_DMADOM_PFN + 1);
    e_dma = min(e_tot, MAX_DMADOM_PFN + 1);
    if ( s_dma < e_dma )
        init_heap_pages(MEMZONE_DMADOM, mfn_to_page(s_dma), e_dma - s_dma);

    s_nrm = max(s_tot, MAX_DMADOM_PFN + 1);
    e_nrm = max(e_tot, MAX_DMADOM_PFN + 1);
    if ( s_nrm < e_nrm )
        init_heap_pages(MEMZONE_DOM, mfn_to_page(s_nrm), e_nrm - s_nrm);
}


struct page_info *alloc_domheap_pages(
    struct domain *d, unsigned int order, unsigned int flags)
{
    struct page_info *pg = NULL;
    cpumask_t mask;
    int i;

    ASSERT(!in_irq());

    if ( !(flags & ALLOC_DOM_DMA) )
    {
        pg = alloc_heap_pages(MEMZONE_DOM, order);
        /* Failure? Then check if we can fall back to the DMA pool. */
        if ( unlikely(pg == NULL) &&
             ((order > MAX_ORDER) ||
              (avail[MEMZONE_DMADOM] <
               (lowmem_emergency_pool_pages + (1UL << order)))) )
            return NULL;
    }

    if ( pg == NULL )
        if ( (pg = alloc_heap_pages(MEMZONE_DMADOM, order)) == NULL )
            return NULL;

    mask = pg->u.free.cpumask;
    tlbflush_filter(mask, pg->tlbflush_timestamp);

    pg->count_info        = 0;
    pg->u.inuse._domain   = 0;
    pg->u.inuse.type_info = 0;

    for ( i = 1; i < (1 << order); i++ )
    {
        /* Add in any extra CPUs that need flushing because of this page. */
        cpumask_t extra_cpus_mask;
        cpus_andnot(extra_cpus_mask, pg[i].u.free.cpumask, mask);
        tlbflush_filter(extra_cpus_mask, pg[i].tlbflush_timestamp);
        cpus_or(mask, mask, extra_cpus_mask);

        pg[i].count_info        = 0;
        pg[i].u.inuse._domain   = 0;
        pg[i].u.inuse.type_info = 0;
        page_set_owner(&pg[i], NULL);
    }

    if ( unlikely(!cpus_empty(mask)) )
    {
        perfc_incrc(need_flush_tlb_flush);
        flush_tlb_mask(mask);
    }

    if ( d == NULL )
        return pg;

    spin_lock(&d->page_alloc_lock);

    if ( unlikely(test_bit(_DOMF_dying, &d->domain_flags)) ||
         unlikely((d->tot_pages + (1 << order)) > d->max_pages) )
    {
        DPRINTK("Over-allocation for domain %u: %u > %u\n",
                d->domain_id, d->tot_pages + (1 << order), d->max_pages);
        DPRINTK("...or the domain is dying (%d)\n", 
                !!test_bit(_DOMF_dying, &d->domain_flags));
        spin_unlock(&d->page_alloc_lock);
        free_heap_pages(pfn_dom_zone_type(page_to_mfn(pg)), pg, order);
        return NULL;
    }

    if ( unlikely(d->tot_pages == 0) )
        get_knownalive_domain(d);

    d->tot_pages += 1 << order;

    for ( i = 0; i < (1 << order); i++ )
    {
        page_set_owner(&pg[i], d);
        wmb(); /* Domain pointer must be visible before updating refcnt. */
        pg[i].count_info |= PGC_allocated | 1;
        list_add_tail(&pg[i].list, &d->page_list);
    }

    spin_unlock(&d->page_alloc_lock);
    
    return pg;
}


void free_domheap_pages(struct page_info *pg, unsigned int order)
{
    int            i, drop_dom_ref;
    struct domain *d = page_get_owner(pg);

    ASSERT(!in_irq());

    if ( unlikely(IS_XEN_HEAP_FRAME(pg)) )
    {
        /* NB. May recursively lock from relinquish_memory(). */
        spin_lock_recursive(&d->page_alloc_lock);

        for ( i = 0; i < (1 << order); i++ )
            list_del(&pg[i].list);

        d->xenheap_pages -= 1 << order;
        drop_dom_ref = (d->xenheap_pages == 0);

        spin_unlock_recursive(&d->page_alloc_lock);
    }
    else if ( likely(d != NULL) )
    {
        /* NB. May recursively lock from relinquish_memory(). */
        spin_lock_recursive(&d->page_alloc_lock);

        for ( i = 0; i < (1 << order); i++ )
        {
            shadow_drop_references(d, &pg[i]);
            ASSERT((pg[i].u.inuse.type_info & PGT_count_mask) == 0);
            pg[i].tlbflush_timestamp  = tlbflush_current_time();
            pg[i].u.free.cpumask      = d->domain_dirty_cpumask;
            list_del(&pg[i].list);
        }

        d->tot_pages -= 1 << order;
        drop_dom_ref = (d->tot_pages == 0);

        spin_unlock_recursive(&d->page_alloc_lock);

        if ( likely(!test_bit(_DOMF_dying, &d->domain_flags)) )
        {
            free_heap_pages(pfn_dom_zone_type(page_to_mfn(pg)), pg, order);
        }
        else
        {
            /*
             * Normally we expect a domain to clear pages before freeing them,
             * if it cares about the secrecy of their contents. However, after
             * a domain has died we assume responsibility for erasure.
             */
            for ( i = 0; i < (1 << order); i++ )
            {
                spin_lock(&page_scrub_lock);
                list_add(&pg[i].list, &page_scrub_list);
                spin_unlock(&page_scrub_lock);
            }
        }
    }
    else
    {
        /* Freeing anonymous domain-heap pages. */
        for ( i = 0; i < (1 << order); i++ )
            pg[i].u.free.cpumask = CPU_MASK_NONE;
        free_heap_pages(pfn_dom_zone_type(page_to_mfn(pg)), pg, order);
        drop_dom_ref = 0;
    }

    if ( drop_dom_ref )
        put_domain(d);
}


unsigned long avail_domheap_pages(void)
{
    unsigned long avail_nrm, avail_dma;

    avail_nrm = avail[MEMZONE_DOM];

    avail_dma = avail[MEMZONE_DMADOM];
    if ( avail_dma > lowmem_emergency_pool_pages )
        avail_dma -= lowmem_emergency_pool_pages;
    else
        avail_dma = 0;

    return avail_nrm + avail_dma;
}



/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/


void let_all_phys_mem_be_mapped () 
{
	int i;

	for ( i = 0; i < e820.nr_map; i++ ) {
		const struct e820_entry *p = &e820.map [ i ];

		if ( p->type != E820_RAM ) {
			continue;
		}

		/* Calculate page-frame range, discarding partial frames. */
		const unsigned long init_mapped_pfn = 1UL << ( 30 - PAGE_SHIFT ); /* 1 GB */
		unsigned long start, end;
		
		start = PFN_UP ( p->addr );
		if ( start < init_mapped_pfn ) {
			start = init_mapped_pfn;
		}
		
		end = PFN_DOWN ( p->addr + p->size );
		if ( end < init_mapped_pfn ) { /* Clip the range to exclude what the bootstrapper initialised. */
			continue;
		}
		
		/* Request the mapping. */
		map_pages_to_xen( VMM_OFFSET + ( start << PAGE_SHIFT ) /* virt */,
				  start /* mfn */, 
				  end - start /* nr_mfn */, 
				  PAGE_HYPERVISOR /* flags */ );
	}
	
	printk("System RAM: %x MB\n",  nr_pages >> ( 20 - PAGE_SHIFT ) );
	
}

static void __init
init_memory_mapping ( )
{
	start_boot_allocator ();
	let_all_phys_mem_be_mapped ();
	
	init_frametable();

	end_boot_allocator();

	/* Initialise the Xen heap, skipping RAM holes. */
	nr_pages = 0;
	for ( i = 0; i < e820.nr_map; i++ )
	{
		if ( e820.map[i].type != E820_RAM )
			continue;
		
		s = e820.map[i].addr;
		e = s + e820.map[i].size;
		if ( s < xenheap_phys_start )
			s = xenheap_phys_start;
		if ( e > xenheap_phys_end )
			e = xenheap_phys_end;
		
		if ( s < e )
		{
			nr_pages += (e - s) >> PAGE_SHIFT;
			init_xenheap_pages(s, e);
		}
	}
	
	printk("Xen heap: %luMB (%lukB)\n", 
	       nr_pages >> (20 - PAGE_SHIFT),
	       nr_pages << (PAGE_SHIFT - 10));

}







