
/*
 * What lives where in the 32-bit address space in the various shadow modes,
 * and what it uses to get/maintain that mapping.
 *
 * SHADOW MODE:      none         enable         translate         external
 *
 * 4KB things:
 * guest_vtable    lin_l2     mapped per gl2   lin_l2 via hl2   mapped per gl2
 * shadow_vtable     n/a         sh_lin_l2       sh_lin_l2      mapped per gl2
 * hl2_vtable        n/a            n/a        lin_hl2 via hl2  mapped per gl2
 * monitor_vtable    n/a            n/a             n/a           mapped once
 *
 * 4MB things:
 * guest_linear  lin via gl2    lin via gl2      lin via hl2      lin via hl2
 * shadow_linear     n/a      sh_lin via sl2   sh_lin via sl2   sh_lin via sl2
 * monitor_linear    n/a            n/a             n/a              ???
 * perdomain      perdomain      perdomain       perdomain        perdomain
 * R/O M2P         R/O M2P        R/O M2P           n/a              n/a
 * R/W M2P         R/W M2P        R/W M2P         R/W M2P          R/W M2P
 * P2M               n/a            n/a           R/O M2P          R/O M2P
 *
 * NB:
 * update_pagetables(), shadow_update_pagetables(), shadow_mode_enable(),
 * shadow_l2_table(), shadow_hl2_table(), and alloc_monitor_pagetable()
 * all play a part in maintaining these mappings.
 */
static void shadow_update_pagetables(struct vcpu *v)
{
	struct domain *d = v->domain;
	unsigned long gmfn = ((v->arch.flags & TF_kernel_mode)?
			      pagetable_get_pfn(v->arch.guest_table) :
			      pagetable_get_pfn(v->arch.guest_table_user));

	unsigned long gpfn = mfn_to_gmfn(d, gmfn);
	unsigned long smfn, old_smfn;

	int need_sync = 0;

	int max_mode = ( shadow_mode_external(d) ? SHM_external
			 : shadow_mode_translate(d) ? SHM_translate
			 : shadow_mode_enabled(d) ? SHM_enable
			 : 0 );

	/*
	 *  arch.guest_vtable
	 */
	if ( max_mode & (SHM_enable | SHM_external ) ) {
		if ( likely(v->arch.guest_vtable != NULL) ) {
			unmap_domain_page_global(v->arch.guest_vtable);
		}
		v->arch.guest_vtable = map_domain_page_global(gmfn);
	}

	/*
	 * Handle 32-bit PAE enabled guest
	 */
	if ( SH_GUEST_32PAE && d->arch.ops->guest_paging_levels == PAGING_L3 ) {
		u32 index = get_cr3_idxval(v);
		gpfn = (index << PGT_score_shift) | gpfn;
	}

	/*
	 *  arch.shadow_table
	 */
	if ( unlikely(!(smfn = __shadow_status(d, gpfn, PGT_base_page_table))) ) {
		smfn = shadow_l4_table(v, gpfn, gmfn);
	} else {
		if ( SH_GUEST_32PAE && d->arch.ops->guest_paging_levels == PAGING_L3 ) {
			update_top_level_shadow(v, smfn);
		}
		/*
		 *  move sync later in order to avoid this smfn been 
		 *  unshadowed occasionally
		 */
		need_sync = 1;
	}


	if ( !get_shadow_ref(smfn) ) {
		BUG();
	}

	old_smfn = pagetable_get_pfn(v->arch.shadow_table);
	v->arch.shadow_table = mk_pagetable((u64)smfn << PAGE_SHIFT);
	if ( old_smfn ) {
		put_shadow_ref(old_smfn);
	}

	SH_VVLOG("shadow_update_pagetables(gmfn=%lx, smfn=%lx)", gmfn, smfn);

	/*
	 * arch.shadow_vtable
	 */
	if ( max_mode == SHM_external || max_mode & SHM_enable )
	{
		if ( v->arch.shadow_vtable ) {
			unmap_domain_page_global(v->arch.shadow_vtable);
		}
		v->arch.shadow_vtable = map_domain_page_global(smfn);
	}

	if(likely(need_sync))
		shadow_sync_all(d);
}

