#include "string.h"
#include "failure.h"
#include "printf.h"
#include "page.h"
#include "alloc.h"


enum pg_table_level {
	PGT_LEVEL_PML4 = 4,
	PGT_LEVEL_PDP  = 3,
	PGT_LEVEL_PD   = 2
};

/* [REF] AMD64 manual Vol. 2, pp. 164-165 */
struct pgt_entry {
	u16 flags: 12; /* Bit 0-11  */
	u64 base:  40; /* Bit 12-52 */
	u16 avail: 11; /* Bit 52-62 */
	u16 nx:    1;  /* Bit 63    */
} __attribute__ ((packed));


static unsigned long 
pg_table_create ( void )
{
	const unsigned long pfn = alloc_pages ( 1, 1 );
	const unsigned long paddr = pfn << PAGE_SHIFT;

	memset ( ( char * ) VIRT ( paddr ), 0, PAGE_SIZE );

	return paddr;
}

unsigned long 
pml4_table_create ( void )
{
	return pg_table_create ( );
}

static unsigned long 
get_index ( unsigned long vaddr, enum pg_table_level level )
{
	unsigned long shift = 0;
	const unsigned long MASK = ( ( 1 << 9 ) - 1 );

	switch ( level ) {
	case PGT_LEVEL_PML4: shift = 39; break;
	case PGT_LEVEL_PDP:  shift = 30; break;
	case PGT_LEVEL_PD:   shift = 21; break;
	default:             fatal_failure ( "wrong level\n" ); break;
	}

	return ( vaddr >> shift ) & MASK;
}

static struct pgt_entry *
get_entry ( unsigned long pg_table_base_vaddr, unsigned long vaddr, enum pg_table_level level )
{
	const unsigned long index = get_index ( vaddr, level );
	return ( struct pgt_entry *) ( pg_table_base_vaddr + index * sizeof ( struct pgt_entry ) );
}

static int 
entry_is_present ( const struct pgt_entry *x )
{
	return x->flags & PTTEF_PRESENT;
}

/* [Note] paging with compatibility-mode (long mode with 2-Mbyte page tranlation) is only supported.  */
static void
__mmap ( unsigned long pg_table_base_vaddr, unsigned long vaddr, unsigned long paddr, enum pg_table_level level )
{
	struct pgt_entry *entry = get_entry ( pg_table_base_vaddr, vaddr, level );

	if ( level == PGT_LEVEL_PD ) {
		/* For page directory entry */

		entry->base = paddr >> PAGE_SHIFT_2MB;
		entry->flags = PTTEF_PRESENT | PTTEF_RW | PTTEF_ACCESSED | PTTEF_DIRTY | PTTEF_PAGE_SIZE | PTTEF_GLOBAL;
		return;
	}

	/* For page-map level-4 entry and page-directory-pointer entry */

	if ( ! entry_is_present ( entry ) ) {
		const unsigned long paddr = pg_table_create ( );
		entry->base  = paddr >> PAGE_SHIFT;
		entry->flags = ( PTTEF_PRESENT | PTTEF_RW | PTTEF_US );
	} 

	// pg_table_base ǻꤵ줿ɥ쥹顤Υ٥ΥڡĴ٤
	const unsigned long next_table_base_vaddr = ( unsigned long ) VIRT ( entry->base );
	__mmap ( next_table_base_vaddr, vaddr, paddr, level - 1 ); 
}

void
mmap ( unsigned long pml4_table_base_vaddr, unsigned long vaddr, unsigned long paddr )
{
	__mmap ( pml4_table_base_vaddr, vaddr, paddr, PGT_LEVEL_PML4 );
}
