#include "string.h"
#include "printf.h"
#include "failure.h"
#include "page.h"
#include "multiboot.h"
#include "system.h"
#include "elf.h"
#include "msr.h"
#include "vm.h"
#include "alloc.h"


static u32
init_vm_mbi ( unsigned long vm_pmem_start )
{
	enum { INSTALL_PADDR = 0x2d0e0UL }; /* < 1 MB */
	struct multiboot_info *mbi = ( struct multiboot_info *) ( vm_pmem_start + INSTALL_PADDR );

	printf ( "Initializing multiboot info.\n" );

	return PHYS ( mbi );
}

static int
is_reserved_pmem ( unsigned long pfn )
{
	return ( pfn >= ( 1 << ( 20 - PAGE_SHIFT ) ) );
}

/* ۥޥʪ¥ޥʪбŤڡơ֥
 * ʪɥ쥹֤*/
static unsigned long 
create_shadow_table ( unsigned long vm_pmem_pfn, unsigned long vm_pmem_size )
{
	unsigned long cr3;
	int i;

	cr3 = pml4_table_create ( );

	for ( i = 0; i < ( vm_pmem_size >> PAGE_SHIFT ); i++ ) {
		const unsigned long pml4     = ( unsigned long ) VIRT ( cr3 );
		const unsigned long vm_paddr = i << PAGE_SHIFT;
		const unsigned long pm_pfn   = i + ( is_reserved_pmem ( i ) ? vm_pmem_pfn : 0 ); /* [NAIVE] for VGA */
		const unsigned long pm_paddr = pm_pfn << PAGE_SHIFT;
		mmap ( pml4, vm_paddr, pm_paddr );
	}

	return cr3;
}

static void
set_registers ( struct vmcb_struct *vmcb )
{
	int i;

	/* Generic Segment Descriptor - Legacy Mode 
	 * (AMD64 manual Vol 2, pp. 98-104) */

	/* setup the segment registers and all their hidden states */
	vmcb->cs.sel = 0xF000;
	vmcb->cs.attributes.bytes = 0x089b; // 1000 1001 1011
	vmcb->cs.limit = 0xffff;
	vmcb->cs.base = 0x000F0000;
	
	// type = 1011 : Data-segment, W(ritable), A(ccessed)
	// S, 
	// P(resent), 
	// G(ranularity)

	struct segment_selector *segregs [] = { &vmcb->ss, &vmcb->ds, &vmcb->es, &vmcb->fs, &vmcb->gs, NULL };
	for ( i = 0; segregs [ i ] != NULL; i++) {
		struct segment_selector *x = segregs [ i ];
		x->sel              = 0x00;
		x->attributes.bytes = 0x0893; // 1000 1001 0011 
		x->limit            = 0xffff;
		x->base             = 0x00;

		// type = 0011 : Code-segment, R(eadable), A(ccessed)
		// S, 
		// P(resent), 
		// G(ranularity)
	}
	
	struct segment_selector *dsegs [] = { &vmcb->ldtr, &vmcb->gdtr, &vmcb->tr, NULL };
	for ( i = 0; dsegs [ i ] != NULL; i++) {
		struct segment_selector *x = dsegs [ i ];
		x->sel              = 0x00;
		x->attributes.bytes = 0x0000;
		x->limit            = 0x0;
		x->base             = 0x00;
	}
	vmcb->idtr.sel = 0x00;
	vmcb->idtr.attributes.bytes = 0x0000;
	vmcb->idtr.limit = 0x3ff;
	vmcb->idtr.base = 0x00;
	
	vmcb->rax = 0;
	vmcb->rsp = 0;

	vmcb->cpl = 0; /* must be equal to SS.DPL  */

	vmcb->cr0 = read_cr0 ( );
//	vmcb->cr3 = pagetable_get_paddr ( svm_paging_enabled(v) ? v->arch.guest_table : ov->domain->arch.phys_table );
	
	unsigned long crn = read_cr4 ( );
	crn &= ~( X86_CR4_PGE | X86_CR4_PSE );
	vmcb->cr4 = crn | X86_CR4_PAE;
}

void
vm_create ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	printf ( "Creating virtual machine.\n" ); 

	/* VMCB Τΰݤ */
	const unsigned long vmcb_pfn = alloc_pages ( 1, 1 );
	struct vmcb_struct *vmcb     = ( struct vmcb_struct * ) VIRT ( vmcb_pfn << PAGE_SHIFT );
	memset ( ( char * ) vmcb, 0, sizeof ( struct vmcb_struct ) );
	vm->vmcb = vmcb;

	/* ۥޥѤʪݤ */
	const unsigned long vm_pmem_pfn = alloc_pages ( vm_pmem_size >> PAGE_SHIFT, 1 );
	unsigned long vm_pmem_start     = ( unsigned long ) VIRT ( vm_pmem_pfn << PAGE_SHIFT );

	/* Copy the OS image to the specified region by interpreting the ELF format.  */
	vmcb->rip = load_elf_image ( guest_image_start, guest_image_size, vm_pmem_start );

#if 0
	/* Setup multiboot info.  */
	vmcb->rbx = init_vm_mbi ( vm_pmem_start );
#endif

	vmcb->h_cr3 = create_shadow_table ( vm_pmem_pfn, vm_pmem_size );

	/* Enable nested paging (See AMD64 manual Vol. 3, p. 488) */
	vmcb->np_enable = 1; 

	/* Flush TLB on VMRUN (all entries, all ASIDs) */
        vmcb->tlb_control = 1;
	
	/* */
	vmcb->guest_asid = 1;

	set_registers ( vmcb );
}

/* [REF] AMD64 manual vol 2, p. 444 */
static int
__check_consistency ( struct vm *vm )
{
	const unsigned long cr0 = read_cr0 ();
	// const unsigned long cr3 = read_cr3 ();
	const unsigned long cr4 = read_cr4 ();

	u32 eax, edx;
	rdmsr ( MSR_EFER, eax, edx );
	const unsigned long efer = ( ( unsigned long ) eax ) | ( ( unsigned long ) edx << 32 );

	if ( ! ( efer & EFER_SVME )  ) {
		return 1;
	}

	if ( ( ! ( cr0 & X86_CR0_PG ) ) && ( cr0 & X86_CR0_NW ) ) {
		return 1;
	}

	if ( ( efer & EFER_LME ) && ( cr0 & X86_CR0_PG ) && ( ! ( cr4 & X86_CR4_PAE ) ) ) {
		return 1;
	}
	
	return 0;
}

static void
check_consistency ( struct vm *vm )
{
	if ( __check_consistency ( vm ) ) {
		fatal_failure ( "Consistency check failed.\n" );
	}
}

void
vm_boot ( struct vm *vm )
{
	printf ( "Booting guest operating system.\n" ); 

	check_consistency ( vm );

	u64 p_vmcb = PHYS ( vm->vmcb );
//	__asm__ __volatile__ ( "movq %0, %%rax" :  : "rm" (rax) );

	/* Switch from VMM to guest OS */
//	__asm__ __volatile__ ( "jmp svm_launch" : : );
	svm_launch ( p_vmcb );

	printf ( "exit_reason = %x\n", vm->vmcb->exitcode );
}

/*

simnow: sys_instr_handlers.cpp:3245: void CProcessor::Handle_VMExit(UINT64, UINT64, UINT64): Assertion `m_CurrentAsid != 0' failed.
롥

  ASID 
   - address space identifier 
   - TLB դ륿

*/
