struct vmcb_struct *alloc_vmcb(void) 
{
	struct vmcb_struct *vmcb = NULL;
	unsigned int order;
	order = get_order_from_bytes(sizeof(struct vmcb_struct)); 

	vmcb = alloc_xenheap_pages(order);
	memset(vmcb, 0, sizeof(struct vmcb_struct));

	return vmcb;
}


/* Set up intercepts to exit the guest into the hypervisor when we want it. */
static int construct_vmcb_controls(struct arch_svm_struct *arch_svm)
{
	struct vmcb_struct *vmcb;
	u32 *iopm;
	u32 *msrpm;

	vmcb = arch_svm->vmcb;

	/* mask off all general 1 intercepts except those listed here */
	vmcb->general1_intercepts = 
		~(GENERAL1_INTERCEPT_CR0_SEL_WRITE | GENERAL1_INTERCEPT_VINTR      | 
		  GENERAL1_INTERCEPT_IDTR_READ     | GENERAL1_INTERCEPT_IDTR_WRITE | 
		  GENERAL1_INTERCEPT_GDTR_READ     | GENERAL1_INTERCEPT_GDTR_WRITE |
		  GENERAL1_INTERCEPT_LDTR_READ     | GENERAL1_INTERCEPT_LDTR_WRITE | 
		  GENERAL1_INTERCEPT_TR_READ       | GENERAL1_INTERCEPT_TR_WRITE   |
		  GENERAL1_INTERCEPT_RDTSC         | GENERAL1_INTERCEPT_PUSHF      |
		  GENERAL1_INTERCEPT_SWINT         | GENERAL1_INTERCEPT_POPF       | 
		  GENERAL1_INTERCEPT_IRET          | GENERAL1_INTERCEPT_PAUSE      |
		  GENERAL1_INTERCEPT_TASK_SWITCH
			);

	/* turn on the general 2 intercepts */
	vmcb->general2_intercepts = 
		GENERAL2_INTERCEPT_VMRUN  | GENERAL2_INTERCEPT_VMMCALL | 
		GENERAL2_INTERCEPT_VMLOAD | GENERAL2_INTERCEPT_VMSAVE  |
		GENERAL2_INTERCEPT_STGI   | GENERAL2_INTERCEPT_CLGI    |
		GENERAL2_INTERCEPT_SKINIT | GENERAL2_INTERCEPT_RDTSCP;

	/* read or write all debug registers 0 - 15 */
	vmcb->dr_intercepts = 0;

	/* RD/WR all control registers 0 - 15, but not read CR2 */
	vmcb->cr_intercepts = ~(CR_INTERCEPT_CR2_READ | CR_INTERCEPT_CR2_WRITE);

	/* The following is for I/O and MSR permision map */
	iopm = alloc_xenheap_pages(get_order_from_bytes(IOPM_SIZE));

	memset(iopm, 0xff, IOPM_SIZE);
	clear_bit(PC_DEBUG_PORT, iopm);
	msrpm = alloc_xenheap_pages(get_order_from_bytes(MSRPM_SIZE));

	memset(msrpm, 0xff, MSRPM_SIZE);

	arch_svm->iopm = iopm;
	arch_svm->msrpm = msrpm;

	vmcb->iopm_base_pa = (u64) virt_to_maddr(iopm);
	vmcb->msrpm_base_pa = (u64) virt_to_maddr(msrpm);

	return 0;
}


/*
 * Initially set the same environement as host.
 */
static int construct_init_vmcb_guest(struct arch_svm_struct *arch_svm, 
                                     struct cpu_user_regs *regs )
{
	int error = 0;
	unsigned long crn;
	segment_attributes_t attrib;
	unsigned long dr7;
	unsigned long eflags;
	unsigned long shadow_cr;
	struct vmcb_struct *vmcb = arch_svm->vmcb;

	/* Allows IRQs to be shares */
	vmcb->vintr.fields.intr_masking = 1;
  
	/* Set up event injection entry in VMCB. Just clear it. */
	vmcb->eventinj.bytes = 0;

	/* TSC */
	vmcb->tsc_offset = 0;
    
	vmcb->cs.sel = regs->cs;
	vmcb->es.sel = regs->es;
	vmcb->ss.sel = regs->ss;
	vmcb->ds.sel = regs->ds; 
	vmcb->fs.sel = regs->fs;
	vmcb->gs.sel = regs->gs;

	/* Guest segment Limits. 64K for real mode*/
	vmcb->cs.limit = GUEST_SEGMENT_LIMIT;
	vmcb->es.limit = GUEST_SEGMENT_LIMIT;
	vmcb->ss.limit = GUEST_SEGMENT_LIMIT;
	vmcb->ds.limit = GUEST_SEGMENT_LIMIT;
	vmcb->fs.limit = GUEST_SEGMENT_LIMIT;
	vmcb->gs.limit = GUEST_SEGMENT_LIMIT;

	/* Base address for segments */
	vmcb->cs.base = 0;
	vmcb->es.base = 0;
	vmcb->ss.base = 0;
	vmcb->ds.base = 0;
	vmcb->fs.base = 0;
	vmcb->gs.base = 0;

	/* Guest Interrupt descriptor table */
	vmcb->idtr.base = 0;
	vmcb->idtr.limit = 0;

	/* Set up segment attributes */
	attrib.bytes = 0;
	attrib.fields.type = 0x3; /* type = 3 */
	attrib.fields.s = 1; /* code or data, i.e. not system */
	attrib.fields.dpl = 0; /* DPL = 0 */
	attrib.fields.p = 1; /* segment present */
	attrib.fields.db = 1; /* 32-bit */
	attrib.fields.g = 1; /* 4K pages in limit */

	/* Data selectors */
	vmcb->es.attributes = attrib; 
	vmcb->ss.attributes = attrib;
	vmcb->ds.attributes = attrib;
	vmcb->fs.attributes = attrib;
	vmcb->gs.attributes = attrib;

	/* Code selector */
	attrib.fields.type = 0xb;   /* type=0xb -> executable/readable, accessed */
	vmcb->cs.attributes = attrib;

	/* Guest Global descriptor table */
	vmcb->gdtr.base = 0;
	vmcb->gdtr.limit = 0;

	/* Guest Local Descriptor Table */
	attrib.fields.s = 0; /* not code or data segement */
	attrib.fields.type = 0x2; /* LDT */
	attrib.fields.db = 0; /* 16-bit */
	attrib.fields.g = 0;   
	vmcb->ldtr.attributes = attrib;

	attrib.fields.type = 0xb; /* 32-bit TSS (busy) */
	vmcb->tr.attributes = attrib;
	vmcb->tr.base = 0;
	vmcb->tr.limit = 0xff;

	__asm__ __volatile__ ("mov %%cr0,%0" : "=r" (crn) :);
	vmcb->cr0 = crn;

	/* Initally PG, PE are not set*/
	shadow_cr = vmcb->cr0;
	shadow_cr &= ~X86_CR0_PG;
	arch_svm->cpu_shadow_cr0 = shadow_cr;

	/* CR3 is set in svm_final_setup_guest */

	__asm__ __volatile__ ("mov %%cr4,%0" : "=r" (crn) :); 
	crn &= ~(X86_CR4_PGE | X86_CR4_PSE | X86_CR4_PAE);
	arch_svm->cpu_shadow_cr4 = crn;
	vmcb->cr4 = crn | SVM_CR4_HOST_MASK;

	vmcb->rsp = 0;
	vmcb->rip = regs->eip;

	eflags = regs->eflags & ~HVM_EFLAGS_RESERVED_0; /* clear 0s */
	eflags |= HVM_EFLAGS_RESERVED_1; /* set 1s */

	vmcb->rflags = eflags;

	__asm__ __volatile__ ("mov %%dr7, %0\n" : "=r" (dr7));
	vmcb->dr7 = dr7;

	return error;
}

int 
construct_vmcb(struct arch_svm_struct *arch_svm, struct cpu_user_regs *regs)
{
	int error;
	long rc=0;

	memset(arch_svm, 0, sizeof(struct arch_svm_struct));

	arch_svm->vmcb = alloc_vmcb();

	/* update the HSA for the current Core */
	set_hsa_to_guest( arch_svm );
	arch_svm->vmcb_pa  = (u64) virt_to_maddr(arch_svm->vmcb);

	construct_vmcb_controls(arch_svm);

	/* guest selectors */
	construct_init_vmcb_guest(arch_svm, regs);


	arch_svm->vmcb->exception_intercepts = MONITOR_DEFAULT_EXCEPTION_BITMAP;
	if (regs->eflags & EF_TF)
		arch_svm->vmcb->exception_intercepts |= EXCEPTION_BITMAP_DB;
	else
		arch_svm->vmcb->exception_intercepts &= ~EXCEPTION_BITMAP_DB;

	return 0;

}
